from numbers import Number
from nptyping import *
from typing import *
from numpy import array, set_printoptions, arange, zeros, interp, mean
from pandas import DataFrame
from copy import deepcopy, copy
from time import process_time

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

import logging
import time
import builtins
set_printoptions(precision=4)
# from .utility_functions import *
from .esbmtk import esbmtkBase, Reservoir, Flux, Signal
from .utility_functions import get_imass, get_frac, get_delta
# from .connections import ConnnectionGroup

class Process(esbmtkBase):
    """This class defines template for process which acts on one or more
     reservoir flux combinations. To use it, you need to create an
     subclass which defines the actual process implementation in their
     call method. See 'PassiveFlux as example'
    
    """
    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """
          Create a new process object with a given process type and options
          """

        self.__defaultnames__()  # default kwargs names
        self.__initerrormessages__()  # default error messages
        self.bem.update({"rate": "a string"})
        self.__validateandregister__(kwargs)  # initialize keyword values
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

    def __postinit__(self) -> None:
        """ Do some housekeeping for the process class
        
        """

        # legacy name aliases
        self.n: str = self.name  # display name of species
        self.r: Reservoir = self.reservoir
        self.f: Flux = self.flux
        self.m: Model = self.r.sp.mo  # the model handle
        self.mo: Model = self.m

        # Create a list of fluxes wich excludes the flux this process
        # will be acting upon
        self.fws: List[Flux] = self.r.lof.copy()
        self.fws.remove(self.f)  # remove this handle

        self.rm0: float = self.r.m[0]  # the initial reservoir mass
        self.direction: Dict[Flux, int] = self.r.lio[self.f]

        self.__misc_init__()

    def __misc_init__(self) -> None:
        """This is just a place holder method which will be called by default
        in __post_init__() This can be overloaded to add additional
        code to the init procedure without the need to redefine
        init. This useful for processes which only define a call method.

        """
        
        pass

    def __defaultnames__(self) -> None:
        """Set up the default names and dicts for the process class. This
          allows us to extend these values without modifying the entire init process

        """

        from .connections import ConnectionGroup
        
        # provide a dict of known keywords and types
        self.lkk: Dict[str, any] = {
            "name": str,
            "reservoir": Reservoir,
            "flux": Flux,
            "rate": Number,
            "delta": Number,
            "lt": Flux,
            "alpha": Number,
            "scale": Number,
            "ref": (Flux, list),
            'register': (str,ConnectionGroup),
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = ["name"]

        # list of default values if none provided
        self.lod: Dict[str, any] = {}

        # default type hints
        self.scale: t
        self.delta: Number
        self.alpha: Number

    def __register__(self, reservoir: Reservoir, flux: Flux) -> None:
        """Register the flux/reservoir pair we are acting upon, and register
          the process with the reservoir
        
        """

        # register the reservoir flux combination we are acting on
        self.f: Flux = flux
        self.r: Reservoir = reservoir
        # add this process to the list of processes acting on this reservoir
        reservoir.lop.append(self)
        flux.lop.append(self)

    def show_figure(self, x, y) -> None:
        """ Apply the current process to the vector x, and show the result as y.
          The resulting figure will be automatically saved.

          Example::
               process_name.show_figure(x,y)
        
          """
        pass

class GenericFunction(Process):
    """This Process class takes a generic function and up to 6 optional
    function arguments, and will replace the mass value(s) of the
    given reservoirs with whatever the function calculates. This is
    particularly useful e.g., to calculate the pH of a given reservoir
    as function of e.g., Alkalinity and DIC.

    Parameters:
     - name = name of process,
     - act_on = name of a reservoir this process will act upon
     - function  = a function reference
     - a1 to a6, up to 6 optional function arguments
    
    in order to use this function we need first declare a function we plan to
    use with the generic function process. This function needs to follow this
    template::

        def my_func(i, a1=0, a2=0, a3=0, a4=0, a5=0, a6=0) -> tuple:
            # 
            # i = index of the current timestep
            # a1 to a2 =  optional function parameter. These must be present, 
            # even if your function will not use it
            
            # calc some stuff and return it as

            return [m, l, h] # where m= mass, and l & h are the respective 
                             # isotopes. If there are none, dummmy values
                             # instead

    
    This function can then be used as::
    
        GenericFunction(name="foo",
                function=my_func,
                a1 = some argument,
                a2 = some argument,
                act_on = reservoir name)
    
    """

    __slots__ = ('function', 'act_on', 'a1', 'a2', 'a3', 'a4', 'a5', 'a6', 'i')

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """
          Create a new process object with a given process type and options
          """

        self.__defaultnames__()  # default kwargs names
        self.lkk: Dict[str, any] = {
            "name": str,
            "act_on": (Flux, Reservoir),
            "function": any,
            "a1": any,
            "a2": any,
            "a3": any,
            "a4": any,
            "a5": any,
            "a6": any,
        }

        # required arguments
        self.lrk: list = (["name", "act_on", "function"])

        # list of default values if none provided
        self.lod: Dict[any, any] = {
            'a1': 0,
            'a2': 0,
            'a3': 0,
            'a4': 0,
            'a5': 0,
            'a6': 0,
        }

        self.__initerrormessages__()  # default error messages
        self.bem.update({
            "act_on": "a reservoir or flux",
            "function": "a function",
            "a1": "a number etc",
            "a2": "a number etc",
            "a3": "a number etc",
            "a4": "a number etc",
            "a5": "a number etc",
            "a6": "a number etc",
        })
        self.__validateandregister__(kwargs)  # initialize keyword values

        if not callable(self.function):
            raise ValueError(
                "function must be defined before it can be used here")

        self.__postinit__()  # do some housekeeping
        self.__register_name__()  #

        # register with reservoir
        if isinstance(self.act_on, Reservoir):
            self.act_on.lpc.append(self)  # register with Reservoir
            self.act_on.mo.lpc_r.append(self)  # Register with Model
        elif isinstance(self.act_on, Flux):
            self.act_on.lpc.append(self)  # register with Flux
            self.act_on.mo.lpc_f.append(self)  # Register with Model
        else:
            raise ValueError(
                "functions can only act upon reservoirs or fluxes")

    def __call__(self, i: int) -> None:
        """Here we execute the user supplied function and assign the 
        return value to the flux or reservoir

        Where i = index of the current timestep
              acting_on = reservoir or flux we are acting on.

        """

        self.act_on[i] = self.function(
            i,
            self.a1,
            self.a2,
            self.a3,
            self.a4,
            self.a5,
            self.a6,
        )

    # redefine post init
    def __postinit__(self) -> None:
        """ Do some housekeeping for the process class

          """

        # legacy name aliases
        self.n: str = self.name  # display name of species
        self.m: Model = self.act_on.sp.mo  # the model handle
        self.mo: Model = self.m

        self.__misc_init__()

class LookupTable(Process):
     """This process replaces the flux-values with values from a static
lookup table

     Example::

     LookupTable("name", upstream_reservoir_handle, lt=flux-object)

     where the flux-object contains the mass, li, hi, and delta values
     which will replace the current flux values.

     """
     
     def __call__(self, r: Reservoir, i: int) -> None:
          """Here we replace the flux value with the value from the flux object 
          which we use as a lookup-table

          """
          self.m[i] :float  = self.lt.m[i]
          self.d[i] :float  = self.lt.d[i]
          self.l[i] :float = self.lt.l[i]
          self.h[i] :float = self.lt.h[i]

class AddSignal(Process):
    """This process adds values to the current flux based on the values provided by the sifnal object.
    This class is typically invoked through the connector object

     Example::

     AddSignal(name = "name",
               reservoir = upstream_reservoir_handle,
               flux = flux_to_act_upon,
               lt = flux with lookup values)

     where - the upstream reservoir is the reservoir the process belongs too
             the flux is the flux to act upon
             lt= contains the flux object we lookup from

    """
    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """
        Create a new process object with a given process type and options
        """

       
        
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["lt", "flux", "reservoir"])  # new required keywords

        self.__initerrormessages__()
        #self.bem.update({"rate": "a string"})
        self.__validateandregister__(kwargs)  # initialize keyword values

        #legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

        # decide whichh call function to use
        if self.mo.m_type == "both":
            self.__execute__ = self.__with_isotopes__
        else:
            self.__execute__ = self.__without_isotopes__

    # setup a placeholder call function
    def __call__(self, r: Reservoir, i: int):
        return self.__execute__(r, i)

    # use this when we do isotopes
    def __with_isotopes__(self, r, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
          the flux value with the value from the signal object which
          we use as a lookup-table (self.lt)

        """
        # add signal mass to flux mass
        self.f.m[i] = self.f.m[i] + self.lt.m[i]
        # add signal delta to flux delta
        self.f.d[i] = self.f.d[i] + self.lt.d[i]

        self.f.l[i], self.f.h[i] = get_imass(self.f.m[i], self.f.d[i], r.rvalue)
        # signals may have zero mass, but may have a delta offset. Thus, we do not know
        # the masses for the light and heavy isotope. As such we have to calculate the masses
        # after we add the signal to a flux

    # use this when we do isotopes
    def __without_isotopes__(self, r, i) -> None:
        """Each process is associated with a flux (self.f). Here we replace
          the flux value with the value from the signal object which
          we use as a lookup-table (self.lt)

        """
        # add signal mass to flux mass
        self.f.m[i] = self.f.m[i] + self.lt.m[i]

class PassiveFlux(Process):
    """This process sets the output flux from a reservoir to be equal to
     the sum of input fluxes, so that the reservoir concentration does
     not change. Furthermore, the isotopic ratio of the output flux
     will be set equal to the isotopic ratio of the reservoir The init
     and register methods are inherited from the process class. The
     overall result can be scaled, i.e., in order to create a split flow etc.
     Example::

     PassiveFlux(name = "name",
                 reservoir = upstream_reservoir_handle
                 flux = flux handle)

     """
    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """ Initialize this Process """

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux"])  # new required keywords
        self.__initerrormessages__()
        #self.bem.update({"rate": "a string"})
        self.__validateandregister__(kwargs)  # initialize keyword values
        #legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """Here we re-balance the flux. That is, we calculate the sum of all fluxes
          excluding this flux. This sum will be equl to this flux. This will likely only
          work for outfluxes though.

        Should this be done for output fluxes as well?
          
          """

        new: float = 0
        for j, f in enumerate(self.fws):
            new += f.m[i] * reservoir.lio[f]

        # self.f[i] = get_flux_data(new,reservoir.d[i-1],reservoir.rvalue)

        m = new
        r = reservoir.l[i - 1] / reservoir.m[i - 1]
        l = m * r
        h = m - l
        self.f.m[i] = m
        self.f.l[i] = l
        self.f.h[i] = h
        self.f.d[i] = reservoir.d[i - 1]

class PassiveFlux_fixed_delta(Process):
     """This process sets the output flux from a reservoir to be equal to
     the sum of input fluxes, so that the reservoir concentration does
     not change. However, the isotopic ratio of the output flux is set
     at a fixed value. The init and register methods are inherited
     from the process class. The overall result can be scaled, i.e.,
     in order to create a split flow etc.  Example::

     PassiveFlux_fixed_delta(name = "name",
                             reservoir = upstream_reservoir_handle,
                             flux handle,
                             delta = delta offset)

     """

     def __init__(self, **kwargs :Dict[str, any]) -> None:
          """ Initialize this Process """


          self.__defaultnames__()  # default kwargs names
          self.lrk.extend(["reservoir","delta", "flux"]) # new required keywords

          self.__initerrormessages__()
          #self.bem.update({"rate": "a string"})
          self.__validateandregister__(kwargs)  # initialize keyword values
          self.__postinit__()  # do some housekeeping

          # legacy names
          self.f :Flux = self.flux
          #legacy variables
          self.mo = self.reservoir.mo

          print("\nn *** Warning, you selected the PassiveFlux_fixed_delta method ***\n ")
          print(" This is not a particularly phyiscal process is this really what you want?\n")
          print(self.__doc__)
          self.__register_name__()
     
     def __call__(self, reservoir :Reservoir, i :int) -> None:
          """Here we re-balance the flux. This code will be called by the
          apply_flux_modifier method of a reservoir which itself is
          called by the model execute method

          """

          r :float = reservoir.rvalue # the isotope reference value

          varflux :Flux = self.f 
          flux_list :List[Flux] = reservoir.lof.copy()
          flux_list.remove(varflux)  # remove this handle

          # sum up the remaining fluxes
          newflux :float = 0
          for f in flux_list:
               newflux = newflux + f.m[i-1] * reservoir.lio[f]

          # set isotope mass according to keyword value
          self.f[i] = array(get_flux_data(newflux, self.delta, r))

class VarDeltaOut(Process):
    """Unlike a passive flux, this process sets the output flux from a
     reservoir to a fixed value, but the isotopic ratio of the output
     flux will be set equal to the isotopic ratio of the reservoir The
     init and register methods are inherited from the process
     class. The overall result can be scaled, i.e., in order to create
     a split flow etc.  Example::

     VarDeltaOut(name = "name",
                 reservoir = upstream_reservoir_handle,
                 flux = flux handle,
                 rate = rate,)

     """

    __slots__ = ('rate')

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """ Initialize this Process
          
          """

        from . import ureg, Q_
        from .connections import ConnectionGroup

        # get default names and update list for this Process
        self.__defaultnames__()
        self.lkk: Dict[str, any] = {
            "name": str,
            "reservoir": Reservoir,
            "flux": Flux,
            "rate": (str, Q_),
            "register": (ConnectionGroup, str),
        }
        self.lrk.extend(["reservoir", "rate"])  # new required keywords
        self.__initerrormessages__()
        self.bem.update({"rate": "a string"})
        self.__validateandregister__(kwargs)  # initialize keyword values

        # parse rate term, and map to legacy name
        self.rateq = Q_(self.rate)
        self.rate = Q_(self.rate).to(self.reservoir.mo.f_unit).magnitude
        #legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """Here we re-balance the flux. This code will be called by the
          apply_flux_modifier method of a reservoir which itself is
          called by the model execute method

        """

        # set flux according to keyword value

        # this explicit expression is siginificantly faster than the below
        # function call
        m = self.rate
        r = reservoir.l[i - 1] / reservoir.m[i - 1]
        l = m * r
        h = m - l
        self.f.m[i] = m
        self.f.l[i] = l
        self.f.h[i] = h
        self.f.d[i] = reservoir.d[i-1]
       
        # self.f[i] = get_flux_data(self.rate,reservoir.d[i-1], reservoir.rvalue)

class ScaleFlux(Process):
    """This process scales the mass of a flux (m,l,h) relative to another
     flux but does not affect delta. The scale factor "scale" and flux
     reference must be present when the object is being initalized

     Example::
          ScaleFlux(name = "Name",
                    reservoir = upstream_reservoir_handle,
                    scale = 1
                    ref = flux we use for scale)

    """

    __slots__ = ('rate', 'scale')
    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """ Initialize this Process 

        """
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux", "scale",
                         "ref"])  # new required keywords

        self.__validateandregister__(kwargs)  # initialize keyword values

        #legacy variables
        self.mo = self.reservoir.mo
        self.__postinit__()  # do some housekeeping
        self.__register_name__()

        

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """Apply the scale factor. This is typically done through the the
          model execute method.
          Note that this will use the mass of the reference object, but that we will set the 
          delta according to the reservoir (or the flux?)

          """
        self.f[i] = self.ref[i] * self.scale
        self.f[i] = get_flux_data(self.f.m[i], reservoir.d[i - 1],
                                  reservoir.rvalue)


class FluxDiff(Process):
    """ The new flux will be the difference of two fluxes

    """
    """This process scales the mass of a flux (m,l,h) relative to another
     flux but does not affect delta. The scale factor "scale" and flux
     reference must be present when the object is being initalized

     Example::
          ScaleFlux(name = "Name",
                    reservoir = upstream_reservoir_handle,
                    scale = 1
                    ref = flux we use for scale)

     """
    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """ Initialize this Process 

        """
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux", "scale",
                         "ref"])  # new required keywords

        self.__validateandregister__(kwargs)  # initialize keyword values
        self.__postinit__()  # do some housekeeping

        #legacy variables
        self.mo = self.reservoir.mo
        self.__register_name__()


    
    def __call__(self, reservoir :Reservoir, i: int) -> None:
        """Apply the scale factor. This is typically done through the the
        model execute method.
        Note that this will use the mass of the reference object, but that we will set the 
        delta according to the reservoir (or the flux?)

        """

        self.f[i] = (self.ref[0][i] - self.ref[1][i]) * self.scale

class Fractionation(Process):
    """This process offsets the isotopic ratio of the flux by a given
        delta value. In other words, we add a fractionation factor

     Example::
          Fractionation(name = "Name",
                        reservoir = upstream_reservoir_handle,
                        flux = flux handle
                        alpha = 12 in permil (e.f)

     """
    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """ Initialize this Process """
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        self.lrk.extend(["reservoir", "flux",
                         "alpha"])  # new required keywords

        self.__validateandregister__(kwargs)  # initialize keyword values
        self.__postinit__()  # do some housekeeping

        # alpha is given in permil, but the fractionation routine expects
        # it as 1 + permil, i.e., 70 permil would 1.007
        #legacy variables
        self.alpha = 1 + self.alpha / 1000
        self.mo = self.reservoir.mo
        self.__register_name__()

        # decide which call function to use
        if self.mo.m_type == "both":
            self.__execute__ = self.__with_isotopes__
        else:
            self.__execute__ = self.__without_isotopes__

    # setup a placeholder call function        
    def __call__(self, reservoir: Reservoir, i: int):
        return self.__execute__(reservoir, i)

    # use this when we do isotopes
    def __with_isotopes__(self, reservoir: Reservoir, i: int) -> None:
        """
        Set flux isotope masses based on fractionation factor

        """

        self.f.l[i], self.f.h[i] = get_frac(self.f.m[i], self.f.l[i],
                                            self.alpha)

        #update delta
        self.f.d[i] = get_delta(self.f.l[i], self.f.h[i], self.f.rvalue)
        return

    # use this when we don't do isotopes
    def __without_isotopes__(self, reservoir: Reservoir, i: int) -> None:
        """
        Set flux isotope masses based on fractionation factor

        """

        return

class RateConstant(Process):
    """This is a wrapper for a variety of processes which depend on rate constants
    Please see the below class definitions for details on how to call them
    At present, the following processes are defined

    ScaleRelativeToNormalizedConcentration
    ScaleRelativeToConcentration

    """
    __slots__ = ('k_value', 'ref_value')
    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """ Initialize this Process

        """

        from . import ureg, Q_
        from .connections import SourceGroup, SinkGroup, ReservoirGroup
        from .connections import ConnectionGroup

        # Note that self.lkk values also need to be added to the lkk
        # list of the connector object.

        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names

        # update the allowed keywords
        self.lkk :dict = {
            "k_value": Number,
            "ref_value": Number,
            "name": str,
            "reservoir": Reservoir,
            "flux": Flux,
            "ref_reservoir": list,
            "left": (list, Reservoir, Number),
            "right": (list, Reservoir, Number),
            "register":
            (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, str),
        }

        # new required keywords
        self.lrk.extend(["reservoir", "k_value"])

        # dict with default values if none provided
        # self.lod = {r

        self.__initerrormessages__()

        # add these terms to the known error messages
        self.bem.update({
            "k_value": "a number",
            "reservoir": "Reservoir handle",
            "ref_reservoirs": "List of Reservoir handle(s)",
            "ref_value": "a number or flux quantity",
            "name": "a string value",
            "flux": "a flux handle",
            "left": "list, reservoir or number",
            "right": "list, reservoir or number",
        })

        # initialize keyword values
        self.__validateandregister__(kwargs)
        self.__postinit__()  # do some housekeeping
        # legacy variables
        self.mo = self.reservoir.mo
        self.__register_name__()


class ScaleRelativeToNormalizedConcentration(RateConstant):
    """This process scales the flux as a function of the upstream
     reservoir concentration C and a constant which describes the
     strength of relation between the reservoir concentration and
     the flux scaling

     F = (C/C0 -1) * k

     where C denotes the concentration in the ustream reservoir, C0
     denotes the baseline concentration and k is a constant
     This process is typically called by the connector
     instance. However you can instantiate it manually as


     ScaleRelativeToNormalizedConcentration(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       K_value =  1000,
                       ref_value = 2 # reference_concentration
    )

    """

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """
          this will be called by the Model.run() method
          """

        scale: float = (reservoir.c[i - 1] / self.ref_value - 1) * self.k_value
        # scale = scale * (scale >= 0)  # prevent negative fluxes.
        self.f[i] = self.f[i] + self.f[i] * array([scale, scale, scale, 1])


class ScaleRelativeToConcentration(RateConstant):
    """This process scales the flux as a function of the upstream
     reservoir concentration C and a constant which describes the
     strength of relation between the reservoir concentration and
     the flux scaling

     F = C * k

     where C denotes the concentration in the ustream reservoir, k is a
     constant. This process is typically called by the connector
     instance. However you can instantiate it manually as


     ScaleRelativeToConcentration(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       K_value =  1000,
    )

    """

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """
          this will be called by the Model.run() method
          """
        #print(f"k= {self.k_value}")
        scale: float = reservoir.c[i - 1] * self.k_value
        # scale = scale * (scale >= 0)  # prevent negative fluxes.
        self.f[i] = self.f[i] * array([scale, scale, scale, 1])


class ScaleRelativeToMass(RateConstant):
    """This process scales the flux as a function of the upstream
     reservoir Mass M and a constant which describes the
     strength of relation between the reservoir mass and
     the flux scaling

     F = F0 *  M * k

     where M denotes the mass in the ustream reservoir, k is a
     constant and F0 is the initial unscaled flux. This process is
     typically called by the connector instance. However you can
     instantiate it manually as

     Note that we scale the flux, rather than compute the flux!

     This is faster than setting a new flux, computing the isotope
     ratio and setting delta. So you either have to set the initial
     flux F0 to 1, or calculate the k_value accordingly

     ScaleRelativeToMass(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       K_value =  1000,
    )

    """

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """
          this will be called by the Model.run() method
          """
        scale: float = reservoir.m[i - 1] * self.k_value
        self.f[i] = self.f[i] * array([scale, scale, scale, 1])


class ScaleRelativeToNormalizedMass(RateConstant):
    """This process scales the flux as a function of the upstream
     reservoir mass M and a constant which describes the
     strength of relation between the reservoir concentration and
     the flux scaling

     F = (M/M0 -1) * k

     where M denotes the mass in the ustream reservoir, M0
     denotes the reference mass, and k is a constant
     This process is typically called by the connector
     instance. However you can instantiate it manually as


     ScaleRelativeToNormalizedConcentration(
                       name = "Name",
                       reservoir= upstream_reservoir_handle,
                       flux = flux handle,
                       K_value =  1,
                       ref_value = 1e5 # reference_mass
    )

    """

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """
          this will be called by the Model.run() method
          """
        scale: float = (reservoir.m[i - 1] / self.ref_value - 1) * self.k_value
        # scale = scale * (scale >= 0)  # prevent negative fluxes.
        self.f[i] = self.f[i] + self.f[i] * array([scale, scale, scale, 1])


class ScaleRelative2otherReservoir(RateConstant):
    """This process scales the flux as a function one or more reservoirs
     constant which describes the
     strength of relation between the reservoir concentration and
     the flux scaling

     F = C1 * C1 * k

     where Mi denotes the concentration in one  or more reservoirs, k is one
     or more constant(s). This process is typically called by the connector
     instance when you specify the connection as

     Connect(source =  upstream reservoir,
               sink = downstream reservoir,
               ctype = "scale_relative_to_multiple_reservoirs"
               ref_reservoirs = [r1, r2, k etc] # you must provide at least one
               k_value = a overall scaling factor
            )
    """

    def __misc_init__(self) -> None:
        """ Test that self.reservoir only contains numbers and reservoirs

        """
        
        for r in self.ref_reservoir:
            if not isinstance(r, (Reservoir)):
                raise ValueError(f"{r} must be reservoir or number, not {type(r)}")

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """
        this will be called by the Model.run() method

        """

        c: float = 1
        for r in self.ref_reservoir:
            c = c * r.c[i - 1]

        scale: float = c * self.k_value
        # scale = scale * (scale >= 0)  # prevent negative fluxes.
        self.f[i] = self.f[i] * array([scale, scale, scale, 1])

class Flux_Balance(RateConstant):
    """This process calculates a flux between two reservoirs as a function
    of multiple reservoir concentrations and constants.

    Note that could result in negative fluxes. which might cause
    issues with isotope ratios (untested)

    This will work with equilibrium reactions between two reservoirs where the
    reaction can be described as

    K * [R1] = R[2] * [R3]

    you can have more than two terms on each side as long as they are
    constants or reservoirs

    Equilibrium(
                name = "Name",
                reservoir = reservoir handle,
                left = [] # list with reservoir names or constants
                right = [] # list with reservoir names or constants
                flux = flux handle,
                k_value = a constant, defaults to 1
    )

    """

    # redefine misc_init which is being called by post-init
    def __misc_init__(self):
        """ Sort out input variables

        """

        Rl: List[Reservoir] = []
        Rr: List[Reservoir] = []
        Cl: List[float] = []
        Cr: List[float] = []
        # parse the left hand side

        em = "left/right values must be constants or reservoirs"
        [self.Rl, self.Cl] = sort_by_type(self.left, [Reservoir, Number], em)
        [self.Rr, self.Cr] = sort_by_type(self.right, [Reservoir, Number], em)

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """
        this will be called by the Model.run() method

        """

        kl: NDArray    = np.array([1.0, 1.0, 1.0, 1.0])
        kr: NDArray    = np.array([1.0, 1.0, 1.0, 1.0])
        scale: NDArray = np.array([1.0, 1.0, 1.0, 1.0])

        # calculate the product of reservoir concentrations for left side
        for r in self.Rl:
            kl *= r[i - 1]
        # multiply with any any constants on the right
        for c in self.Cl:
            kl *= c

        # calculate the product of reservoir concentrations for right side
        for r in self.Rr:
            kr *= r[i - 1]
        # multiply with any any constants on the right
        for c in self.Cr:
            kr *= c

        # set flux
        self.f[i] = (kl - kr) *  self.k_value

class Monod(Process):
    """This process scales the flux as a function of the upstream
     reservoir concentration using a Michaelis Menten type
     relationship

     F = F * a * F0 x C/(b+C)

     where F0 denotes the unscaled flux (i.e., at t=0), C denotes
     the concentration in the ustream reservoir, and a and b are
     constants.

     Example::
          Monod(name = "Name",
                reservoir =  upstream_reservoir_handle,
                flux = flux handle ,
                ref_value = reference concentration
                a_value = constant,
                b_value = constant )

     """

    def __init__(self, **kwargs: Dict[str, any]) -> None:
        """

        """

        from . import ureg, Q_

        """ Initialize this Process """
        # get default names and update list for this Process
        self.__defaultnames__()  # default kwargs names
        
        # update the allowed keywords
        self.lkk :dict = {
            "a_value": Number,
            "b_value": Number,
            "ref_value": (Number,str, Q_),
            "name": str,
            "reservoir": Reservoir,
            "flux": Flux,
            "register":
            (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, str),
        }

        self.lrk.extend(["reservoir", "a_value", "b_value",
                         "ref_value"])  # new required keywords

        self.__initerrormessages__()
        self.bem.update({
            "a_value": "a number",
            "b_value": "a number",
            "reservoir": "Reservoir handle",
            "ref_value": "a number",
            "name": "a string value",
            "flux": "a flux handle",
        })

        self.__validateandregister__(kwargs)  # initialize keyword values
        self.__postinit__()  # do some housekeeping
        #legacy variables
        self.mo = self.reservoir.mo
        self.__register_name__()

    def __call__(self, reservoir: Reservoir, i: int) -> None:
        """
          this willbe called by Model.execute apply_processes
          """

        scale: float = self.a_value * (self.ref_value * reservoir.c[i - 1]) / (
            self.b_value + reservoir.c[i - 1])

        scale = scale * (scale >= 0)  # prevent negative fluxes.
        self.f[i] + self.f[i] * scale

    def __plot__(self, start: int, stop: int, ref: float, a: float,
                 b: float) -> None:
        """ Test the implementation

          """

        y = []
        x = range(start, stop)

        for e in x:
            y.append(a * ref * e / (b + e))

        fig, ax = plt.subplots()  #
        ax.plot(x, y)
        # Create a scatter plot for ax
        plt.show()



"""
     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# from pint import UnitRegistry
from numbers import Number
from nptyping import *
from typing import *
from numpy import array, set_printoptions, arange, zeros, interp, mean
from pandas import DataFrame
from copy import deepcopy, copy
from time import process_time
#from numba import jit, njit

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import mpmath

import logging
import time
import builtins
import math
set_printoptions(precision=4)

def get_imass(m: float, d: float, r: float) -> [float, float]:
    """
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio 
    species
    
    """

    li: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    hi: float = m - li
    return [li, hi]


def get_frac(m: float, l: float, a: float) -> [float, float]:
    """Calculate the effect of the istope fractionation factor alpha on
    the ratio between the light and heavy isotope.

    """

    li: float = -l * m / (a * l - a * m - l)
    hi: float = m - li  # get the new heavy isotope value
    return li, hi


def get_flux_data(m: float, d: float, r: float) -> [NDArray, float]:
    """ 
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio 
    species. Unlike get_mass, this function returns the full array
    
    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    h: float = m - l

    return np.array([m, l, h, d])

def get_delta(l: [NDArray, [Float64]], h: [NDArray, [Float64]],
              r: float) -> [NDArray, [Float64]]:
    """Calculate the delta from the mass of light and heavy isotope
     Arguments are l and h which are the masses of the light and
     heavy isotopes respectively, r = abundance ratio of the
     respective element. Note that this equation can result in a
     siginificant loss of precision (on the order of 1E-13). I
     therefore round the results to numbers largers 1E12 (otherwise a
     delta of zero may no longer be zero)

   """

    d: float = 1E3 * (abs(h) / abs(l) - r) / r
    return d

def add_to (l, e):
    """
      add element e to list l, but check if the entry already exist. If so, throw
      exception. Otherwise add
    """

    if not (e in l):  # if not present, append element
        l.append(e)

def get_plot_layout(obj):
      """ Simple function which selects a row, column layout based on the number of
      objects to display.  The expected argument is a reservoir object which
      contains the list of fluxes in the reservoir

      """

      noo = 1 # the reservoir is the fisrt object
      for f in obj.lof: # count numbert of fluxes
            if f.plot == "yes":
                  noo += 1
                  
      for d in obj.ldf: # count number of data fields
            noo +=1
            
      # noo = len(obj.lof) + 1  # number of objects in this reservoir
      logging.debug(f"{noo} subplots for {obj.n} required")

      if noo < 2:
          geo = [1, 1]  # one row, one column
          size = [5, 3]  # size in inches
      elif 1 < noo < 3:
          geo = [2, 1]  # two rows, one column
          size = [5, 6]  # size in inches
      elif 2 < noo < 5:
          geo = [2, 2]  # two rows, two columns
          size = [10, 6]  # size in inches
      elif 4 < noo < 7:
          geo = [2, 3]  # two rows, three columns
          size = [15, 6]  # size in inches
      elif 6 < noo < 10:
          geo = [3, 3]  # two rows, three columns
          size = [15, 9]  # size in inches
      else:
          print("plot geometry for more than 8 fluxes is not yet defined")
          print("Consider calling flux.plot individually on each flux in the reservoir")
          # print(f"Selected Geometry: rows = {geo[0]}, cols = {geo[1]}")

      return size, geo

def list_fluxes(self,name,i) -> None:
            """
            Echo all fluxes in the reservoir to the screen
            """
            print(f"\nList of fluxes in {self.n}:")
            
            for f in self.lof: # show the processes
                  direction = self.lio[f.n]
                  if direction == -1:
                        t1 = "From:"
                        t2 = "Outflux from"
                  else:
                        t1 = "To  :"   
                        t2 = "Influx to"

                  print(f"\t {t2} {self.n} via {f.n}")
                  
                  for p in f.lop:
                        p.describe()

            print(" ")
            for f in self.lof:
                  f.describe(i) # print out the flux data

def show_data(self, **kwargs) -> None:
    """ Print the 3 lines of the data starting with index

    Optional arguments:
    
    index :int = 0 starting index
    indent :int = 0 indentation 
    """

    off: str = "  "

    if "index" not in kwargs:
        index = 0
    else:
        index = kwargs["index"]

    if "indent" in kwargs:
        ind: str = kwargs["indent"] * " "
    else:
        ind: str = ""

    # show the first 4 entries
    for i in range(index, index + 3):
        print(
            f"{off}{ind}i = {i}, Mass = {self.m[i]:.2e}, delta = {self.d[i]:.2f}"
        )

def set_y_limits(ax: plt.Axes, model: any) -> None:
    """ Prevent the display or arbitrarily small differences
    """
    lower: float
    upper: float

    bottom, top = ax.get_ylim()
    if (top - bottom) < model.display_precision:
        top = bottom + model.display_precision
        ax.set_ylim(bottom, top)


def get_ptype(obj, kwargs: dict) -> int:
    """
    Set plot type variable
    
    """

    ptype: int = 0
    if "ptype" in kwargs:
        if kwargs["ptype"] == "both":
            ptype = 0
        elif kwargs["ptype"] == "iso":
            ptype = 1
        elif kwargs["ptype"] == "concentration":
            ptype = 2
        elif kwargs["ptype"] == "mass_only":
            ptype = 2
    else:
        if obj.m_type == "mass_only":
            ptype = 2
        elif obj.m_type == "both":
            ptype = 0
        else:
            raise ValueError(
                "ptype must be one of 'both/iso/concentration/mass_only'")

    return ptype


def plot_object_data(geo: list, fn: int, obj, ptype: int) -> None:
    """collection of commands which will plot and annotate a reservoir or flux
      object into an existing plot window. 
      """

    from . import ureg, Q_
    from esbmtk import Flux, Reservoir, Signal, DataField

    # geo = list with rows and cols
    # fn  = figure number
    # yl  = array with y values for the left side
    # yr  = array with y values for the right side
    # obj = object handle, i.e., reservoir or flux

    first_axis: bool = False
    second_axis: bool = False

    rows = geo[0]
    cols = geo[1]
    # species = obj.sp
    model = obj.mo
    time = model.time + model.offset

    # convert data from model units to display units (i.e. the same
    # units the input data was defined).
    # time units are the same regardless of object
    time = (time * model.t_unit).to(model.d_unit).magnitude

    # we do not map isotope values
    yr = obj.d

    # remap concentration & flux values
    if isinstance(obj, Flux):
        yl = (obj.m * model.f_unit).to(obj.plt_units).magnitude
        y_label = f"{obj.legend_left} [{obj.plt_units:~P}]"

    elif isinstance(obj, Reservoir):
        if obj.display_as == "mass":
            yl = (obj.m * model.m_unit).to(obj.plt_units).magnitude
            y_label = f"{obj.legend_left} [{obj.plt_units:~P}]"

        elif obj.transform_m != "None":
            if callable(obj.transform_m):
                #yl = (obj.m * model.m_unit).to(obj.plt_units).magnitude
                yl = obj.transform_m(obj.m)
                y_label = f"{obj.legend_left}"
            else:
                raise ValueError("transform_m must be function")

        else:
            yl = (obj.c * model.c_unit).to(obj.plt_units).magnitude
            y_label = f"{obj.legend_left} [{obj.plt_units:~P}]"

    elif isinstance(obj, Signal):
        # use the same units as the associated flux
        yl = (obj.c * model.c_unit).to(obj.fo.plt_units).magnitude
        y_label = f"{obj.n} [{obj.fo.plt_units:~P}]"

    elif isinstance(obj, DataField):
        time = (time * model.t_unit).to(model.d_unit).magnitude
        yl = obj.y1_data
        y_label = obj.y1_label
        if obj.y2_data == "None":
            ptype = 2
        else:
            ptype = 0

    else:  # sources, sinks, external data should not show up here
        raise ValueError(f"{obj.n} = {type(obj)}")

    # decide what to plot
    if ptype == 0:
        first_axis = True
        second_axis = True
    elif ptype == 1:
        first_axis = False
        second_axis = True
    elif ptype == 2:
        first_axis = True
        second_axis = False

    # start subplot
    ax1 = plt.subplot(rows, cols, fn, title=obj.n)

    # set color index
    cn = 0
    col = f"C{cn}"

    if first_axis:
        # plot left y-scale data
        ln1 = ax1.plot(time[1:-2], yl[1:-2], color=col, label=obj.legend_left)
        # set labels
        ax1.set_xlabel(f"{model.time_label} [{model.d_unit:~P}]")
        ax1.set_ylabel(y_label)
        # remove unnecessary frame species
        ax1.spines['top'].set_visible(False)
        set_y_limits(ax1, model)

    # set color index
    cn = cn + 1
    col = f"C{cn}"

    if second_axis:
        ax2 = ax1.twinx()  # create a second y-axis

        # plof right y-scale data
        ln2 = ax2.plot(time[1:-2], yr[1:-2], color=col, label=obj.legend_right)

        ax2.set_ylabel(obj.ld)  # species object delta label
        ax2.spines['top'].set_visible(
            False)  # remove unnecessary frame speciess
        set_y_limits(ax2, model)

    # adjust display properties for title and legend
    ax1.set_title(obj.n)
    plt.rcParams['axes.titlepad'] = 14  # offset title upwards
    plt.rcParams["legend.facecolor"] = '0.8'  # show a gray background
    plt.rcParams["legend.edgecolor"] = '0.8'  # make frame the same color
    plt.rcParams["legend.framealpha"] = 0.4  # set transparency

    for d in obj.led:  # loop over external data objects if present

        if isinstance(d.x[0], str):  # if string, something is off
            raise ValueError("No time axis in external data object {d.name}")
        if "y" in dir(d):  # mass or concentration data is present
            cn = cn + 1
            col = f"C{cn}"
            leg = f"{obj.lm} {d.legend}"
            ln3 = ax1.scatter(d.x, d.y, color=col, label=leg)
        if "z" in dir(d) and second_axis:  # isotope data is present
            cn = cn + 1
            col = f"C{cn}"
            leg = f"{d.legend}"
            ln3 = ax2.scatter(d.x, d.z, color=col, label=leg)

    # collect all labels and print them in one legend
    if first_axis:
        handler1, label1 = ax1.get_legend_handles_labels()

    if second_axis:
        handler2, label2 = ax2.get_legend_handles_labels()

    if first_axis and second_axis:
        legend = ax2.legend(handler1 + handler2, label1 + label2,
                            loc=0).set_zorder(6)
    #elif first_axis:
    #    legend = ax1.legend(handler1 + label1, loc=0).set_zorder(6)
    #elif second_axis:
    #   legend = ax2.legend(handler2 + label2, loc=0).set_zorder(6)

    # Matplotlib will show arbitrarily small differences which can be confusing
    #yl_min = min(yl)
    #yl_max = max(yl)
    #if (yl_max - yl_min) < 0.1:

def is_name_in_list(n: str, l: list) -> bool:
    """ Test if an object name is part of the object list
    
    """

    r: bool = False
    for e in l:
        if e.n == n:
            r = True
    return r


def get_object_from_list(n: str, l: list) -> any:
    """ Match a name to a list of objects. Return the object
    
    """

    for o in l:
        if o.n == n:
            r = o
    return r

def get_hplus(dic :float, ta :float)->float:
    """
    Calculate H+ concentration based on DIC concentration and Alkalinity
    according to eq 11 in Follows et al 2006
    
    """

    pk1 = 5.81  # at this ph value CO2 and HCO3 have the same concentration
    pk2 = 8.92
    K1 = 10**-pk1
    K2 = 10**-pk2
    
    g = dic / ta
    hplus = 0.5 * ((g - 1) * K1 + ((1 - g)**2 * K1**2 - 4 * K1 * K2 *
                                   (1 - 2 * g))**0.5)

    return hplus

def get_pco2(dic :float, ta :float) -> float:
    """Calculate pCO2 in uatm at 25C and a Salinity of 35

    DIC has to be in mmol/l!

    """
    pk1 = 5.81  # at this ph value CO2 and HCO3 have the same concentration
    pk2 = 8.92
    K1 = 10**-pk1
    K2 = 10**-pk2
    K0 = 36

    hplus = get_hplus(dic,ta)

    # get [CO2] in water
    co2 = dic / (1 + K1/hplus + K1*K2/hplus**2)

    # get pco2 as a function of co2 fugacity
    pco2 = co2/K0 *1E6

    # this cam also be expressed in teh following way
    #pco2a = (ta/K0 * ( K1/hplus + 2*K1*K2/hplus**2)**-1) * 1.e6
    #pco2b = (dic/K0 * (1 + K1/hplus + (K1*K2)/hplus**2)**-1) * 1.e6
   
    return pco2

def sort_by_type(l: list, t: list, m: str) -> list:
    """divide a list by type into new lists. This function will return a
    list and it is up to the calling code to unpack the list

    l is list with various object types
    t is a list which contains the object types used for sorting
    m is a string for the error function
    """

    #from numbers import Number

    lc = l.copy()
    rl = []

    for ot in t:  # loop over object types
        a = []
        for e in l:  # loop over list elements
            if isinstance(e, ot):
                a.append(e)  # add to temporary list
                lc.remove(e)  # remove this element

        rl.append(a)  # save the temporary list to rl

    # at this point, all elements of lc should have been processed
    # if not, lc contains element which are of a different type
    if len(lc) > 0:
        raise TypeError(m)

    return rl

def get_string_between_brackets(s :str) -> str:
    """ Parse string and extract substring between square brackets

    """
    
    s =  s.split("[")
    if len(s) < 2:
        raise ValueError(f"Column header {s} must include units in square brackets")

    s = s[1]

    s = s.split("]")

    if len(s) < 2:
        raise ValueError(f"Column header {s} must include units in square brackets")

    return s[0]

def map_units(v: any, *args) -> float:
    """ parse v to see if it is a string. if yes, map to quantity. 
        parse v to see if it is a quantity, if yes, map to model units
        and extract magnitude, assign mangitude to return value
        if not, assign value to return value
        
        v : a keyword value number/string/quantity
        args: one or more quantities (units) see the Model class (e.g., f_unit)

    """

    from . import Q_

    m: float = 0
    match :bool = False

    # test if string, map to quantity if yes
    if isinstance(v, str):
        v = Q_(v)

    # test if we find a matching dimension, map if true
    if isinstance(v, Q_):
        for q in args:
            if v.dimensionality == q.dimensionality:
                m = v.to(q).magnitude
                match = True

        if not match:
            message = f"{v} is none of {print(*args)}"
            raise ValueError(message)

    else:  # no quantity, so it should be a number
        m = v

    if not isinstance(m, Number):
        raise ValueError(f"m is {type(m)}, must be float, v={v}. Something is fishy")

    return m



"""

     esbmtk.connections
     ~~~~~~~~~~~~~~~~~~

     Classes which handle the connections and fluxes between esbmtk objects
     like Reservoirs, Sources, and Sinks.

     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# from pint import UnitRegistry
from __future__ import annotations
from numbers import Number
from nptyping import *
from typing import *
from numpy import array, set_printoptions, arange, zeros, interp, mean
from pandas import DataFrame
from copy import deepcopy, copy
from time import process_time

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

import logging
import time
import builtins
set_printoptions(precision=4)
from .utility_functions import map_units
from .processes import *
from .esbmtk import *

class Connect(esbmtkBase):
    """Two reservoirs connect to each other via at least one flux. This
    module creates the connecting flux and creates a connector object
    which stores all connection properties.

    For simple connections, the type flux type is derived implcitly from the specified parameters.
    For complex connections, the flux type must be set explicitly. See the examples below:

    Parameters:
        - name: A string which determines the name of this object. Optional, if not provided
          the connection name will be derived as "Source_2_Sink_Connection"
        - source: An object handle for a Source or Reservoir
        - sink: An object handle for a Sink or Reservoir
        - rate: A quantity (e.g., "1 mol/s"), optional
        - delta: The isotope ratio, optional
        - alpha: A fractionation factor, optional
        - id: A string wich will become part of the object name, optional
        - plot: "yes" or "no", defaults to "yes"
        - signal: An object handle of signal, optional
        - pl: A list of process objects, optional
        - ctype: connection type, optional, this allows to scale a flux in response to other
          reservoirs and fluxes

    Connection Types:
    -----------------
    Basic Connections (the advanced ones are below):

    - If both =rate= and =delta= are given, the flux is treated as a
       fixed flux with a given isotope ratio. This is usually the case for
       most source objects (they can still be affected by a signal, see
       above), but makes little sense for reservoirs and sinks.

    - If both the =rate= and =alpha= are given, the flux rate is fixed
      (subject to any signals), but the isotopic ratio of the output
      flux depends on the isotopic ratio of the upstream reservoir
      plus any isotopic fractionation specified by =alpha=. This is
      typically the case for fluxes which include an isotopic
      fractionation (i.e., pyrite burial). This combination is not
      particularly useful for source objects.

    - If the connection specifies only =delta= the flux is treated as a
      variable flux which is computed in such a way that the reservoir
      maintains steady state with respect to it's mass.

    - If the connection specifies only =rate= the flux is treated as a
      fixed flux which is computed in such a way that the reservoir
      maintains steady state with respect to it's isotope ratio.

    Examples of Basic Connections
    -----------------------------

    Connecting a Source to a Reservoir
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Unless you use a Signal, a source typically provides a steady stream with a given isotope ratio (if used)
    
    Example::

       Connect(source =  Source,
               sink = downstrean reservoir,
               rate = "1 mol/s",
               delta = optional,
               signal = optional, see the signal documentation)
    
    Connecting a Reservoir to Sink or another Reservoir
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Here we can distinguish between cases where we use fixed flux, or a flux which reacts to in some way to the 
    upstream reservoir (see the Reservoir to Reservoir section for a more complete treatment):

    Fixed outflux, with no isotope fractionation

    Example::
    
         Connect(source =  upstream reservoir,
               sink = Sink,
               rate = "1 mol/s",)

    Fixed outflux, with isotope fractionation

    Example::
    
         Connect(source =  upstream reservoir,
               sink = Sink,
               alpha = -28,
               rate = "1 mol/s",)

    Advanced Connections
    --------------------

    You can aditionally define connection properties via the ctype
    keyword. This requires additional keyword parameters. The following values are
    recognized

    ctype = "scale_with_flux"
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    This will scale a flux relative to another flux:

    Example::
     
        Connect(source =  upstream reservoir,
               sink = downstream reservoir,
               ctype = "scale_with_flux",
               ref = flux handle, 
               k_value = a scaling factor)

    ctype = "scale_with_mass" and "scale_with_concentration"
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    This will scale a flux relative to the mass or concentration of a reservoir

    Example::
     
        Connect(source =  upstream reservoir,
               sink = downstream reservoir,
               ctype = "scale_with_mass",
               ref = reservoir handle, 
               k_value = a scaling factor)

    ctype = "scale_relative_to_multiple_reservoirs"
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    This process scales the flux as a function one or more reservoirs
    or constants which describes the
    strength of relation between the reservoir concentrations and
    the flux scaling

    F = C1 * C2 * k

    where Ci denotes the concentrartion in one  or more reservoirs, k is one
    or more constants. 

    Example:: 

        Connect(source =  upstream reservoir,
               sink = downstream reservoir,
               ctype = "scale_relative_to_multiple_reservoirs"
               ref_reservoirs = [r1, r2, k etc] # you must provide at least one
               k_value = a scaling factor)
    
    K_value is an overall scaling factor.


    ctype = "flux_balance"
    ~~~~~~~~~~~~~~~~~~~~~~

   This type can be used to express equilibration fluxes 
   between two reservoirs. This connection type, takes three parameters:

   - =left= is a list which can contain constants and/or reservoirs. The
     list must contain at least one valid element. All elements in this
     list will be multiplied with each other. E.g. if we have a list
     with one constant and one reservoir, the reservoir concentration
     will be multiplied with the constant. If we have two reservoirs,
     the respective reservoir concentrations will be multiplied with
     each other.
   - =right= similar to =left= The final flux rate will be computed as
     the difference between =left= and =right=
   - =k_value= a constant which will be multiplied with the difference
     between =left=and =right=

    Example::

        Connect(source=R_CO2,         # target of flux
                sink=R_HCO3,          # source of flux
                rate="1 mol/s",       # flux rate
                ctype="flux_balance", # connection type 
                k_value=1,            # global scaling factor
                left=[K1, R_CO2],     # where K1 is a constant
                right=[R_HCO3, R_Hplus])

    
    Useful methods in this class
    ----------------------------
    The following methods might prove useful
    
     - describe() will provide a short description of the connection objects.
     - list_processes() which will list all the processes which are associated with this connection.
     - update() which allows you to update connection properties after the connection has been created

    """
    def __init__(self, **kwargs):
        """ The init method of the connector obbjects performs sanity checks e.g.:
               - whether the reservoirs exist
               - correct flux properties (this will be handled by the process object)
               - whether the processes do exist (hmmh, that implies that the optional processes do get registered with the model)
               - creates the correct default processes
               - and connects the reservoirs

        see the class documentation for details and examples

        """

        from . import ureg, Q_

        # provide a dict of all known keywords and their type
        self.lkk: Dict[str, any] = {
            "name":
            str,
            "id":
            str,
            "source": (Source, Reservoir),
            "sink": (Sink, Reservoir),
            "delta": (Number, str),
            "rate": (str, Number, Q_),
            "pl":
            list,
            "alpha": (Number, str),
            "species":
            Species,
            "ctype":
            str,
            "ref": (Flux, list),
            "react_with":
            Flux,
            "ratio":
            Number,
            "scale":
            Number,
            "ref_value": (str, Number, Q_),
            "ref_reservoir": (list, Reservoir),
            "k_value": (Number, str, Q_),
            "a_value":
            Number,
            "b_value":
            Number,
            "left": (list, Number, Reservoir),
            "right": (list, Number, Reservoir),
            "plot":
            str,
            "register":
            (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, str),
            "signal": (Signal, str),
        }

        if "name" not in kwargs:
            n = kwargs["source"].name + "_2_" + kwargs[
                "sink"].n + "_Connector"  # set the name
            kwargs.update({"name": n})  # and add it to the kwargs

        # provide a list of absolutely required keywords
        self.lrk: list = ["name", "source", "sink"]

        # list of default values if none provided
        self.lod: Dict[any, any] = {
            "id": "None",
            "plot": "yes",
            "ctype": "None",
            "delta": "None",
            "alpha": "None",
            "rate": "None",
            "k_value": 1,
            "signal": "None",
        }

        # validate and initialize instance variables
        self.__initerrormessages__()

        self.bem.update({
            "k_concentration": "a number",
            "k_mass": "a number",
            "k_value": "a number",
            "a_value": "a number",
            "ref_value": "a number, string, or quantity",
            "b_value": "a number",
            "name": "a string",
            "id": "a string",
            "plot": "a string",
            "left": "Number, list or Reservoir",
            "right": "Number, list or Reservoir",
            "signal": "Signal Handle",
        })

        self.drn = {
            "alpha": "_alpha",
            "rate": "_rate",
            "delta": "_delta",
        }

        self.__validateandregister__(kwargs)

        if kwargs["id"] != "None":
            self.name = self.name + f"_{self.id}"
        if 'pl' in kwargs:
            self.lop: list[Process] = self.pl
        else:
            self.lop: list[Process] = []

        if self.signal != "None":
            self.lop.append(self.signal)

        # if no reference reservoir is specified, default to the upstream
        # reservoir
        if 'ref_reservoir' not in kwargs:
            self.ref_reservoir = kwargs["source"]

        # legacy names
        self.influx: int = 1
        self.outflux: int = -1
        self.n = self.name
        self.mo = self.source.sp.mo
        self.p = 0  # the default process handle
        self.r1: (Process, Reservoir) = self.source
        self.r2: (Process, Reservoir) = self.sink

        self.get_species(self.r1, self.r2)  #
        self.mo: Model = self.sp.mo  # the current model handle
        self.lof: list[Flux] = []  # list of fluxes in this connection
        # get a list of all reservoirs registered for this species
        self.lor: list[Reservoir] = self.mo.lor

        self.source.loc.add(self)  # register connector with reservoir
        self.sink.loc.add(self)  # register connector with reservoir
        self.mo.loc.add(self)  # register connector with model

        self.__create_flux__()  # Source/Sink/Regular

        self.__set_process_type__()  # derive flux type and create flux(es)

        self.__register_name__()  # register connection in namespace

        # This should probably move to register fluxes
        self.__register_process__()

        if self.register == "yes":
            print(f"Created connection {self.name}")
        else:
            print(f"Created group connection {self.register.name}.{self.name}")

    def update(self, **kwargs):
        """Update connection properties. This will delete existing processes
        and fluxes, replace existing key-value pairs in the
        self.kwargs dict, and then re-initialize the connection.

        """
        self.__delete_process__()
        self.__delete_flux__()
        self.kwargs.update(kwargs)
        self.__init_connection__(self.kwargs)
        print(f"Updated {self.n}")

    def get_species(self, r1, r2) -> None:
        """In most cases the species is set by r2. However, if we have
        backward fluxes the species depends on the r2

        """
        #print(f"r1 = {r1.n}, r2 = {r2.n}")
        if isinstance(self.r1, Source):
            self.r = r1
        else:  # in this case we do have an upstream reservoir
            self.r = r2

        # test if species was explicitly given
        if "species" in self.kwargs:  # this is a quick fix only
            self.sp = self.kwargs["species"]
        else:
            self.sp = self.r.sp  # get the parent species

    def __create_flux__(self) -> None:
        """Create flux object, and register with reservoir and global namespace

        """

        # test if default arguments present
        if self.delta == "None":
            d = 0
        else:
            d = self.delta

        if self.rate == "None":
            r = f"1 {self.sp.mo.f_unit}"
            #self._rate = r
        else:
            r = self.rate

        # flux name
        if self.id == "None":
            n = self.r1.n + '_2_' + self.r2.n + "_Flux"
        else:
            n = self.r1.n + '_2_' + self.r2.n + "_" + \
                self.id + "_Flux"  # flux name r1_2_r2

        # derive flux unit from species obbject
        funit = self.sp.mu + "/" + str(self.sp.mo.bu)  # xxx

        self.fh = Flux(
            name=n,  # flux name
            species=self.sp,  # Species handle
            delta=d,  # delta value of flux
            rate=r,  # flux value
            plot=self.plot,  # display this flux?
            register=self.register,  # is this part of a group?
        )

        # register flux with its reservoirs
        if isinstance(self.r1, Source):
            # add the flux name direction/pair
            self.r2.lio[self.fh] = self.influx
            # add the handle to the list of fluxes
            self.r2.lof.append(self.fh)
            # register flux and element in the reservoir.
            self.__register_species__(self.r2, self.r1.sp)

        elif isinstance(self.r2, Sink):
            # add the flux name direction/pair
            self.r1.lio[self.fh] = self.outflux
            # add flux to the upstream reservoir
            self.r1.lof.append(self.fh)
            # register flux and element in the reservoir.
            self.__register_species__(self.r1, self.r2.sp)

        elif isinstance(self.r1, Sink):
            raise NameError(
                "The Sink must be specified as a destination (i.e., as second argument"
            )

        elif isinstance(self.r2, Source):
            raise NameError("The Source must be specified as first argument")

        else:  # this is a regular connection
            # add the flux name direction/pair
            self.r1.lio[self.fh] = self.outflux
            # add the flux name direction/pair
            self.r2.lio[self.fh] = self.influx
            # add flux to the upstream reservoir
            self.r1.lof.append(self.fh)
            # add flux to the downstream reservoir
            self.r2.lof.append(self.fh)
            self.__register_species__(self.r1, self.r1.sp)
            self.__register_species__(self.r2, self.r2.sp)

        self.lof.append(self.fh)

    def __register_species__(self, r, sp) -> None:
        """ Add flux to the correct element dictionary"""
        # test if element key is present in reservoir
        if sp.eh in r.doe:
            # add flux handle to dictionary list
            r.doe[sp.eh].append(self.fh)
        else:  # add key and first list value
            r.doe[sp.eh] = [self.fh]

    def __register_process__(self) -> None:
        """ Register all flux related processes"""

        # first test if we have a signal in the list. If so,
        # remove signal and replace with process

        p_copy = copy(self.lop)
        for p in p_copy:  # loop over process list if provided during init
            if isinstance(p, Signal):
                self.lop.remove(p)
                if p.ty == "addition":
                    # create AddSignal Process object
                    n = AddSignal(name=p.n + "_addition_process",
                                  reservoir=self.r,
                                  flux=self.fh,
                                  lt=p.data)
                    self.lop.append(n)
                else:
                    raise ValueError(f"Signal type {p.ty} is not defined")

        # nwo we can register everythig on lop
        for p in self.lop:
            p.__register__(self.r, self.fh)

    def __set_process_type__(self) -> None:
        """ Deduce flux type based on the provided flux properties. The method calls the
        appropriate method init routine
        """

        if isinstance(self.r1, Source):
            self.r = self.r2
        else:
            self.r = self.r1

        # set process name
        #if len(self.kwargs["id"]) > 0:
        if self.id == "None":
            self.pn = self.r1.n + "_2_" + self.r2.n
        else:
            self.pn = self.r1.n + "_2_" + self.r2.n + f"_{self.id}"

        # if connection type is not set explicitly
        if self.ctype == "None":
            # set the fundamental flux type based on the flux arguments given
            if self.delta != "None" and self.rate != "None":
                pass  # do nothing but exit this conditional
            # variable flux with fixed delta
            elif self.delta != "None":  # rate must None
                self.__passivefluxfixeddelta__()
            elif self.rate != "None":  # delta must be None
                self.__vardeltaout__()  # variable delta with fixed flux
            else:  # if neither are given -> default varflux type
                self._delta = 0
                self.__passiveflux__()
                print("passive flux")

        elif self.ctype == "flux_diff":
            self.__vardeltaout__()
            self.__flux_diff__()
        elif self.ctype == "scale_with_flux":
            self.__vardeltaout__()
            self.__scaleflux__()
        elif self.ctype == "copy_flux":
            self.__vardeltaout__()
            self.__scaleflux__()
        elif self.ctype == "scale_with_mass":
            self.__rateconstant__()
        elif self.ctype == "scale_with_concentration":
            self.__rateconstant__()
        elif self.ctype == "scale_with_concentration_normalized":
            self.__rateconstant__()
        elif self.ctype == "scale_with_mass_normalized":
            self.__rateconstant__()
        elif self.ctype == "scale_relative_to_multiple_reservoirs":
            self.__rateconstant__()
        elif self.ctype == "flux_balance":
            self.__rateconstant__()
        elif self.ctype == "monod_type_limit":
            self.__vardeltaout__()
            self.__rateconstant__()
        else:
            print(f"Connection Type {self.type} is unknown")
            raise ValueError(f"Unknown connection type {self.ctype}")

        # Set optional flux processes
        if self.alpha != "None":
            self.__alpha__()  # Set optional flux processes

    def __passivefluxfixeddelta__(self) -> None:
        """ Just a wrapper to keep the if statement manageable

        """

        ph = PassiveFlux_fixed_delta(
            name=self.pn + "_Pfd",
            reservoir=self.r,
            flux=self.fh,
            register=self.register,
            delta=self.delta)  # initialize a passive flux process object
        self.lop.append(ph)

    def __vardeltaout__(self) -> None:
        """Unlike a passive flux, this process sets the output flux from a
        reservoir to a fixed value, but the isotopic ratio of the
        output flux will be set equal to the isotopic ratio of the
        upstream reservoir.

        """

        ph = VarDeltaOut(name=self.pn + "_Pvdo",
                         reservoir=self.r,
                         flux=self.fh,
                         register=self.register,
                         rate=self.rate)
        self.lop.append(ph)

    def __scaleflux__(self) -> None:
        """ Scale a flux relative to another flux

        """

        if not isinstance(self.kwargs["ref"], Flux):
            raise ValueError("Scale reference must be a flux")

        ph = ScaleFlux(name=self.pn + "_PSF",
                       reservoir=self.r,
                       flux=self.fh,
                       register=self.register,
                       scale=self.kwargs["k_value"],
                       ref=self.kwargs["ref"])
        self.lop.append(ph)

    def __flux_diff__(self) -> None:
        """ Scale a flux relative to the difference between
        two fluxes

        """

        if not isinstance(self.kwargs["ref"], list):
            raise ValueError("ref must be a list")

        ph = FluxDiff(name=self.pn + "_PSF",
                      reservoir=self.r,
                      flux=self.fh,
                      register=self.register,
                      scale=self.kwargs["k_value"],
                      ref=self.kwargs["ref"])
        self.lop.append(ph)

    def __reaction__(self) -> None:
        """ Just a wrapper to keep the if statement manageable

        """

        if not isinstance(self.kwargs["react_with"], Flux):
            raise ValueError("Scale reference must be a flux")
        ph = Reaction(name=self.pn + "_RF",
                      reservoir=self.r,
                      flux=self.fh,
                      register=self.register,
                      scale=self.kwargs["ratio"],
                      ref=self.kwargs["react_with"])
        # we need to make sure to remove the flux referenced by
        # react_with is removed from the list of fluxes in this
        # reservoir.
        self.r2.lof.remove(self.kwargs["react_with"])
        self.lop.append(ph)

    def __passiveflux__(self) -> None:
        """ Just a wrapper to keep the if statement manageable

        """

        ph = PassiveFlux(
            name=self.pn + "_PF",
            reservoir=self.r,
            register=self.register,
            flux=self.fh)  # initialize a passive flux process object
        self.lop.append(ph)  # add this process to the process list

    def __alpha__(self) -> None:
        """ Just a wrapper to keep the if statement manageable

        """

        ph = Fractionation(name=self.pn + "_Pa",
                           reservoir=self.r,
                           flux=self.fh,
                           register=self.register,
                           alpha=self.kwargs["alpha"])
        self.lop.append(ph)  #

    def __rateconstant__(self) -> None:
        """ Add rate constant type process

        """

        from . import ureg, Q_

        # this process requires that we use the vardeltaout process
        if self.mo.m_type != "mass_only":
            self.__vardeltaout__()

        if self.ctype == "scale_with_mass":
            self.k_value = map_units(self.k_value, self.mo.m_unit)
            ph = ScaleRelativeToMass(name=self.pn + "_PkM",
                                     reservoir=self.ref_reservoir,
                                     flux=self.fh,
                                     register=self.register,
                                     k_value=self.k_value)

        elif self.ctype == "scale_with_mass_normalized":
            self.k_value = map_units(self.k_value, self.mo.m_unit)
            self.ref_value = map_units(self.ref_value, self.mo.m_unit)
            ph = ScaleRelativeToNormalizedMass(name=self.pn + "_PknM",
                                               reservoir=self.ref_reservoir,
                                               flux=self.fh,
                                               register=self.register,
                                               ref_value=self.ref_value,
                                               k_value=self.k_value)

        elif self.ctype == "scale_with_concentration":
            self.k_value = map_units(self.k_value, self.mo.c_unit,
                                     self.mo.f_unit, self.mo.r_unit)
            ph = ScaleRelativeToConcentration(name=self.pn + "_PkC",
                                              reservoir=self.ref_reservoir,
                                              flux=self.fh,
                                              register=self.register,
                                              k_value=self.k_value)

        elif self.ctype == "scale_relative_to_multiple_reservoirs":
            self.k_value = map_units(self.k_value, self.mo.c_unit,
                                     self.mo.f_unit, self.mo.r_unit)
            ph = ScaleRelative2otherReservoir(name=self.pn + "_PkC",
                                              reservoir=self.source,
                                              ref_reservoir=self.ref_reservoir,
                                              flux=self.fh,
                                              register=self.register,
                                              k_value=self.k_value)

        elif self.ctype == "flux_balance":
            self.k_value = map_units(self.k_value, self.mo.c_unit,
                                     self.mo.f_unit, self.mo.r_unit)
            ph = Flux_Balance(name=self.pn + "_Pfb",
                              reservoir=self.source,
                              left=self.left,
                              right=self.right,
                              flux=self.fh,
                              register=self.register,
                              k_value=self.k_value)

        elif self.ctype == "scale_with_concentration_normalized":
            self.k_value = map_units(self.k_value, self.mo.c_unit,
                                     self.mo.f_unit, self.mo.r_unit)
            self.ref_value = map_units(self.ref_value, self.mo.c_unit)
            ph = ScaleRelativeToNormalizedConcentration(
                name=self.pn + "_PknC",
                reservoir=self.ref_reservoir,
                flux=self.fh,
                register=self.register,
                ref_value=self.ref_value,
                k_value=self.k_value)

        elif self.ctype == "monod_ctype_limit":
            self.ref_value = map_units(self.ref_value, self.mo.c_unit)
            ph = Monod(name=self.pn + "_PMonod",
                       reservoir=self.ref_reservoir,
                       flux=self.fh,
                       register=self.register,
                       ref_value=self.ref_value,
                       a_value=self.a_value,
                       b_value=self.b_value)

        else:
            raise ValueError(
                f"This should not happen,and points to a keywords problem in {self.name}"
            )

        self.lop.append(ph)

    def describe(self, **kwargs) -> None:
        """ Show an overview of the object properties.
        Optional arguments are
        index  :int = 0 this will show data at the given index
        indent :int = 0 indentation

        """
        off: str = "  "
        if "index" not in kwargs:
            index = 0
        else:
            index = kwargs["index"]

        if "indent" not in kwargs:
            indent = 0
            ind = ""
        else:
            indent = kwargs["indent"]
            ind = ' ' * indent

        # print basic data bout this Connection
        print(f"{ind}{self.__str__(indent=indent)}")

        print(f"{ind}Fluxes:")
        for f in sorted(self.lof):
            f.describe(indent=indent, index=index)

    def __delete_process__(self) -> None:
        """ Updates to the connection properties may change the connection type and thus
        the processes which are associated with this connection. We thus have to
        first delete the old processes, before we re-initialize the connection

        """

        # identify which processes we need to delete
        # unregister process from connection.lop, reservoir.lop, flux.lop, model.lmo
        # delete process from global name space if present

        lop = copy(self.lop)
        for p in lop:
            for f in lof:
                if isinstance(f.register, ConnectionGroup):
                    # remove from Connection group list of model objects
                    self.register.lmo.remove(f)
                else:
                    self.r1.lop.remove(p)
                    self.fh.lop.remove(p)
                    self.lop.remove(p)
                    self.r1.mo.lmo.remove(p.n)
                    del p

    def __delete_flux__(self) -> None:
        """ Updates to the connection properties may change the connection type and thus
        the processes which are associated with this connection. We thus have to
        first delete the old flux, before we re-initialize the connection

        """

        # identify which processes we need to delete
        # unregister process from connection.lop, reservoir.lop, flux.lop, model.lmo
        # delete process from global name space if present

        lof = copy(self.lof)
        for f in lof:
            if isinstance(f.register, ConnectionGroup):
                # remove from Connection group list of model objects
                self.register.lmo.remove(f)
            else:
                self.r1.lof.remove(f)
                self.lof.remove(f)
                self.r1.mo.lmo.remove(f.n)
                del f

    # ---- Property definitions to allow for connection updates --------
    """ Changing the below properties requires that we delete all
    associated objects (processes), and determines the new flux type,
    and initialize/register these with the connection and model.
    We also have to update the keyword arguments as these are used
    for the log entry

    """

    # ---- alpha ----
    @property
    def alpha(self) -> Number:
        return self._alpha

    @alpha.setter
    def alpha(self, a: Number) -> None:
        self.__delete_process__()
        self.__delete_flux__()
        self._alpha = a
        self.kwargs["alpha"] = a
        self.__set_process_type__()  # derive flux type and create flux(es)
        self.__register_process__()

    # ---- rate  ----
    @property
    def rate(self) -> Number:
        return self._rate

    @rate.setter
    def rate(self, r: str) -> None:
        from . import ureg, Q_
        self.__delete_process__()
        self.__delete_flux__()
        self._rate = Q_(r).to(self.mo.f_unit)
        self.kwargs["rate"] = r
        self.__create_flux__()  # Source/Sink/Regular
        self.__set_process_type__()  # derive flux type and create flux(es)
        self.__register_process__()

    # ---- delta  ----
    @property
    def delta(self) -> Number:
        return self._delta

    @delta.setter
    def delta(self, d: Number) -> None:
        self.__delete_process__()
        self.__delete_flux__()
        self._delta = d
        self.kwargs["delta"] = d
        self.__create_flux__()  # Source/Sink/Regular
        self.__set_process_type__()  # derive flux type and create flux(es)
        self.__register_process__()

class Connection(Connect):
    """ Alias for the Connect class

    """

class ConnectionGroup(esbmtkBase):
    """Name:

        ConnectionGroup

        Connect reservoir/sink/source groups when at least one of the
        arguments is a reservoirs_group object. This method will
        create regular connections for each matching species. 

        Use the connection.update() method to fine tune connections 
        after creation

    Example::

        ConnectionGroup(source =  upstream reservoir / upstream reservoir group
           sink = downstrean reservoir / downstream reservoirs_group
           delta = defaults to zero and has to be set manually
           alpha =  defaults to zero and has to be set manually
           rate = shared between all connections
           ref = shared between all connections
           species = list, optional, if present, only these species will be connected
           ctype = if set it will be shared between all connections. To
           pl = [list]) process list. optional, shared between all connections
           id = optional identifier, shared between all connections
           plot = "yes/no" # defaults to yes, shared between all connections
        )

        Notes: if species is given as a list, shared arguments like, delta, alpha, rate, ref,
        ctype pl, and plot can also be provided as list. As long as there is a one to one mapping
        the species list and the list of a shared property, the shared property will be mapped
        to each species, e.g.:

        species = [CO, Hplus]
        alpha = [1.02, 1.03]

        will create two connections, the first one with an alpha of 1.02, and the second with an alpha of 1.03

    """
    def __init__(self, **kwargs) -> None:

        # provide a dict of all known keywords and their type
        self.lkk: Dict[str, any] = {
            "id": dict,
            "name": str,
            "source": (SourceGroup, ReservoirGroup),
            "sink": (SinkGroup, ReservoirGroup),
            "delta": dict,
            "rate": dict,
            "pl": dict,
            "signal": Signal,
            "alpha": dict,
            "species": dict,
            "ctype": str,
            "ref": list,
            "plot": dict,
        }

        self.base_name = kwargs["source"].name + "_2_" + kwargs["sink"].name

        n = kwargs["source"].name + "_2_" + kwargs[
            "sink"].name + "_ConnectionGroup"  # set the name

        # set connection group name
        kwargs.update({"name": n})  # and add it to the kwargs

        # provide a list of absolutely required keywords
        self.lrk: list = ["source", "sink"]

        # get the number of sub reservoirs in the source and sink
        nor_sink = len(kwargs["sink"].species)
        nor_source = len(kwargs["source"].species)

        #if nor_source != nor_sink:
        #    raise ValueError(
        #        "Number of sub reservoirs does not match. Specify match explicitly"
        #    )

        cid: dict = {}
        plot: dict = {}
        delta: dict = {}
        alpha: dict = {}
        rate: dict = {}
        # loop over names and create dicts
        for n in kwargs['sink'].species:
            cid[n] = 'None'
            plot[n] = 'no'
            delta[n] = 'None'
            alpha[n] = 'None'
            rate[n] = 'None'

        # list of default values if none provided
        self.lod: Dict[any, any] = {
            "id": cid,
            "plot": plot,
            "delta": delta,
            "alpha": alpha,
            "rate": rate
        }

        # turn kwargs into instance variables
        self.__validateandregister__(kwargs)

        self.loc: list = []  # list of connections in this group

        # self.source.lor is a  list with the object names in the group
        self.mo = self.sink.lor[0].mo

        # loop over sub-reservoirs and create connections
        for i, r in enumerate(self.source.lor):
            if not isinstance(r, (Reservoir, Source, Sink)):
                raise ValueError(
                    f"{r} must be of type reservoir, source or sink")
                # take the species of this sub reservoir
                # in the source, and find matching
                # species in the sink

            # loop over sink list until a match is found
            for j, s in enumerate(self.sink.lor):
                if not isinstance(s, (Reservoir, Source, Sink)):
                    raise ValueError(
                        f"{r} must be of type reservoir, source or sink")

                if r.species == s.species:  # match found
                    # name = f"{self.source.name}_{r.species.name}_2_{self.sink.name}_{s.species.name}"
                    name = f"{r.species.name}_2_{s.species.name}_Connector"
                    a = Connect(
                        name=name,
                        source=r,
                        sink=s,
                        rate=self.rate[s.species],
                        delta=self.delta[s.species],
                        alpha=self.alpha[s.species],
                        plot=self.plot[s.species],
                        id=self.id[s.species],
                        register=self,
                    )
                elif j == nor_sink:  # no match was found
                    raise ValueError("{r.species} has no match")

            # register connection with connection group
            #this should happen automatically
            #setattr(self, a.name, a)
            self.loc.append(a)

        # register connection group in global namespace
        self.__register_name__()

    def describe(self) -> None:
        """ List all connections in this group
        
        """

        print(f"Group Connection from {self.source.name} to {self.sink.name}\n")
        print("    The following Connections are part of this group\n")

        for c in self.loc:
            print(c.name)

        print("")

        print(f"        You can query the details of each connection like this:\n")
        print(f"          {self.name}.{self.loc[0].name}.describe()")



from __future__ import annotations

from pint import UnitRegistry

ureg = UnitRegistry(on_redefinition="ignore")
Q_ = ureg.Quantity

ureg.define("Sverdrup = 1e6 * meter **3 / second = Sv = Sverdrups")
ureg.define("Mol = 1 * mol / liter = M")

# import utility_functions
# import esbmtk

# from .base_class import esbmtkBase


# rom .species_definitions import carbon, sulfur, hydrogen, phosphor
from .esbmtk import *
from .connections import ConnectionGroup, Connection, Connect
from .utility_functions import *
